// Copyright (C) 2006-2013 Xilinx, Inc.  All rights reserved
// $Id: DeviceModel.h,v 1.12.206.17 2013/07/11 01:02:35 jcorbett Exp $
#ifndef Iv_DeviceModel_h
#define Iv_DeviceModel_h

#include <string>
#include <cstring>
#include <set>
#include <algorithm>

#include "Portability/export/Port_Type.h"
#include "Pds/Pah/export/Pds_PahDesignLoaderT.h"
#include "Databases/Rf/export/Facade/Rf_Device.h"
#include "Databases/Rf/export/Facade/Rf_Package.h"
#include "Databases/Rf/export/Facade/Rf_CellDef.h"
#include "Databases/Rf/export/Facade/Rf_Cell.h"
#include "Databases/Rf/export/Facade/Rf_Pin.h"
#include "Databases/Rf/export/Facade/Rf_PinContainer.h"
#include "Databases/Rf/export/Facade/Rf_Node.h"
#include "Databases/Rf/export/Facade/Rf_Pip.h"
#include "Databases/Rf/export/Facade/Rf_Wire.h"
#include "Databases/Rf/export/Facade/Rf_WireContainer.h"
#include "Databases/Rf/export/Rf_TilePinHelpers.h"
#include "TUtilities/export/regex.hpp"

#include "main.h"

class PM_PIC;

class Tile;

// Functor for accumulating the bounding box of a collection of Rf_Cells.
class BoundInts {
public:
   BoundInts() : initialized(false), tile(0) {};
   void operator()(Rf_Cell cell) {
      ASSERT(cell);
      Port::UINT32 cellCol(cell.GetTileColumn());
      Port::UINT32 cellRow(cell.GetTileRow());
      if (initialized) {
         if (cellCol               < left.GetTileColumn())  {   left = cell; }
         if (right.GetTileColumn() < cellCol)               {  right = cell; }
         if (cellRow               < bottom.GetTileRow())   { bottom = cell; }
         if (top.GetTileRow()      < cellRow)               {    top = cell; }
      } else {
         left = right = top = bottom = cell;
         initialized = true;
      }
   }
   bool initialized;
   Rf_Cell left;
   Rf_Cell top;
   Rf_Cell right;
   Rf_Cell bottom;
   Rf_Cell site;
   Tile* tile;
};

// Functor for accumulating the canonical corners of a collection of Rf_Cells.
class CornerInts {
public:
   CornerInts() : initialized(false) {};
   void operator()(Rf_Cell cell) {
      ASSERT(cell);
      Port::UINT32 cellCol(cell.GetTileColumn());
      Port::UINT32 cellRow(cell.GetTileRow());
      if (initialized) {
         if ((cellCol < left && cellRow >= bottom) ||
             (cellCol == left && cellRow > bottom))
         {
            lowerLeft = cell;
            left = cellCol;
            bottom = cellRow;
         } else if ((right < cellCol && top >= cellRow) ||
                    (cellCol == right && top > cellRow))
         {
            upperRight = cell;
            top = cellRow;
            right = cellCol;
         }
      } else {
         lowerLeft = upperRight = cell;
         left = right = cell.GetTileColumn();
         top = bottom = cell.GetTileRow();
         initialized = true;
      }
   }
   bool initialized;
   size_t left;
   size_t right;
   size_t top;
   size_t bottom;
   Rf_Cell lowerLeft;
   Rf_Cell upperRight;
};

// Class to load and query a device model; and to construct the
// low-level data structures used in tile-based isolation analysis
// (code involving SiteRect).
class DeviceModel {
public:
   static const boost::regex configCenterPat;
   static const boost::regex clockBufferPat;
public:
   typedef struct SiteRect {
      std::string areaGroupName;
      std::string siteType;
      std::string rangeString;
      Rf_Cell lowerCell;
      Rf_Cell upperCell;
      Rf_Cell lowerInt;
      Rf_Cell upperInt;
      bool canonical;          // lower left of upper

      // UCF site names do not always match device model site
      // names. For 7 series and Spartan-6, the difference can be
      // accounted for with stemming and a couple of explicit
      // exceptions.
      bool UcfSiteEqual(const std::string& deviceModelSite) const {
         // UCF site names do not always match device model site
         // names. The following sites are related to their UCF
         // counterparts by the removal of a suffix. The patterns
         // below match device model names. The captured expression is
         // the corresponding UCF name.
         //
         // FIFO18E1 and RAMBFIFO36E1 are treated as special cases
         // because they map to RAMB18 and RAMB36 respectively.
         static const boost::regex
            siteSuffixPat("(DSP48|IDELAY|ILOGIC|IOB|ODELAY|"
                          "OLOGIC|PCIE|RAMB18|SLICE)"
                          "(?:_2_1|_A1|A1|BWER|E1|E2|E2_FINEDELAY|E3|"
                          "IBRATE|L|M|S|SE|X|"
                          "18|18M|18S|2|33|33M|33S)");
         boost::smatch result;
         if (boost::regex_match(deviceModelSite, result, siteSuffixPat)) {
            return siteType == result.str(1);
         } else if (deviceModelSite == "FIFO18E1") {
            return siteType == "RAMB18";
         } else if (deviceModelSite == "RAMBFIFO32E1") {
            return siteType == "RAMB36";
         } else {
            return siteType == deviceModelSite;
         }
      }

      // Simple geometric containment ignoring the fact that many area
      // range rectangles are not densely populated, e.g. a range that
      // spans two DSP columns will have columns of slices in between.
      bool Contains(Port::UINT32 column, Port::UINT32 row) const {
         Port::UINT32 x0(lowerInt.GetTileColumn());
         // Note y-coordinate inversion.
         Port::UINT32 y0(upperInt.GetTileRow());
         Port::UINT32 x1(upperInt.GetTileColumn());
         Port::UINT32 y1(lowerInt.GetTileRow());

         return y0 <= row && row <= y1 &&
            ((canonical && x0 <= column  && column <= x1) ||
             (x1 <= column && column <= x0));
      };
      // Is a site of the given type in the rectangle at the given
      // point? 
      bool Contains(Port::UINT32 column, Port::UINT32 row,
                    const std::string& deviceModelSite) const {
         std::string sitePrefix(siteType, 0, 4);
         if (sitePrefix == "PHAS") {
            sitePrefix = siteType.substr(0, sizeof "PHASER_x");
         }
         return IsPrefix(sitePrefix.c_str(), deviceModelSite.c_str()) &&
            Contains(column, row);
      };
      // Used for debugging.
      std::string ToString() const {
         return areaGroupName + " " + siteType + " " + rangeString + " (" +
            ::ToString(lowerCell.GetTileColumn()) + "," + ::ToString(lowerCell.GetTileRow()) + ") (" +
            ::ToString(upperCell.GetTileColumn()) + "," + ::ToString(upperCell.GetTileRow()) + ")";
      };
   } SiteRect;
   class SiteRectLess {
   public:
      bool operator()(const SiteRect& a,const SiteRect& b) const {
         if (a.areaGroupName < b.areaGroupName) return true;
         if (b.areaGroupName < a.areaGroupName) return false;
         if (a.siteType < b.siteType) return true;
         if (b.siteType < a.siteType) return false;
         Rf_Cell aLower(a.lowerCell);
         Rf_Cell bLower(b.lowerCell);
         if (aLower.GetTileColumn() < bLower.GetTileColumn()) return true;
         if (bLower.GetTileColumn() < aLower.GetTileColumn()) return false;
         if (aLower.GetTileRow()    < bLower.GetTileRow())    return true;
         if (bLower.GetTileRow()    < aLower.GetTileRow())    return false;
         Rf_Cell aUpper(a.upperCell);
         Rf_Cell bUpper(b.upperCell);
         if (aUpper.GetTileColumn() < bUpper.GetTileColumn()) return true;
         if (bUpper.GetTileColumn() < aUpper.GetTileColumn()) return false;
         if (aUpper.GetTileRow()    < bUpper.GetTileRow())    return true;
         if (bUpper.GetTileRow()    < aUpper.GetTileRow())    return false;
         if (a.rangeString          < b.rangeString)          return true;
         return false;
      };
   };
   class SiteRectEqual {
   public:
      bool operator()(const SiteRect& a,const SiteRect& b) const{
         bool result(!SiteRectLess()(a, b) && !SiteRectLess()(b, a));
         return result;
      };
   };
   typedef std::vector<SiteRect> SiteRectVec;
   typedef SiteRectVec::iterator SiteRectVecIt;
   class CellCoordLess {
   public:
      bool operator()(const Rf_Cell& a,const Rf_Cell& b) const {
         if (a.GetTileRow()    < b.GetTileRow())    return true;
         if (b.GetTileRow()    < a.GetTileRow())    return false;
         if (a.GetTileColumn() < b.GetTileColumn()) return true;
         if (b.GetTileColumn() < a.GetTileColumn()) return false;
         return std::string(a.GetName()) < std::string(b.GetName());
      };
   };
   typedef std::set<Rf_Cell, CellCoordLess> CellSet;
   typedef CellSet::iterator CellSetIt;
public:
   DeviceModel(const std::string deviceName,
               const std::string packageName);
   DeviceModel(const Rf_Device* device);
   ~DeviceModel();
   const Rf_Device* GetRfDevice() const;
   const Rf_Package* GetRfPackage() const;
   const std::string GetPartName() const;
   const std::string GetArchitectureName() const;
   Port::UINT32 GetPackageColumnCount() const;
   Port::UINT32 GetPackageRowCount() const;
   Port::UINT32 GetTileColumnCount() const;
   Port::UINT32 GetTileRowCount() const;
   Rf_CellContainer GetCells() const;
   Rf_CellContainer GetSites() const;
   Rf_PackagePin GetPackagePin(Rf_Cell cell) const;
   Rf_Cell GetSiteParentTile(Rf_Cell site) const;
   bool IsUserTileType(const std::string& siteType) const;
   bool IsShortableSuperSite(const Rf_Cell superSite) const;
   bool IsConfigCenterType(const std::string& siteType) const;
   bool IsInterconnectCell(const Rf_Cell cell) const;
   bool Is7SeriesClockSite(Rf_Cell site) const;
   bool IsPhaserSafePin(Rf_Pin pin) const;
   // Trace from each pin on site until a branch or dead end is reached.
   // A branch usually indicates an INT. When an INT is found, call
   // cellFn() on the INT cell.
   template <typename CellFn>
   void TraceToInt(CellFn& cellFn, Rf_Cell site) const {
      std::vector<Rf_Node> prevNodes;
      prevNodes.reserve(20);

      // The following CMT pins do not carry user data and are
      // excluded to make it possible to run a horizontal fence across
      // a 7-series CMT column.
      // Note: this function is only used for 7-series CMT components.
      Rf_PinContainer sitePins(site.GetPins());
      // For each pin, trace until a branch
      for (Rf_PinContainer::const_iterator pIt(sitePins.begin());
           pIt != sitePins.end();
           ++pIt)
      {
         const Rf_Pin pin(*pIt);
         const bool isSafePin(IsPhaserSafePin(pin));
         Rf_Node node(pin.GetNode());

         // Trace until a branch point
         bool gotBranch(false);
         // Collect the path upto the first branch in prevNodes.
         prevNodes.push_back(node);
         int dummyCount(20);    // Used to guarantee termination.
         do {
            std::vector<Rf_Node> nodes;
            nodes.reserve(20);
            const Port::UINT32
               nodeCount(node.GetOtherNodes(Rf_Node::All, nodes));
            if (nodeCount == 0) {
               break;
            }
            // Find the branch point
            gotBranch = (2U - node.IsPin() < nodeCount);
            if (gotBranch) {
               break;
            }
            if (nodes.size() == 1) {
               if (node == nodes.front()) {
                  break;        // Path looped back to origin without branching
               }
               node = nodes.front();
            } else if (nodes.size() == 2) {
               // Want the element of nodes that is not alread in prevNodes.
               node =
                  std::count(prevNodes.begin(), prevNodes.end(), nodes.at(0))
                  ? nodes.at(1) : nodes[0];
            } else {
               gotBranch = true;
            }
            if (std::count(prevNodes.begin(), prevNodes.end(), node)) {
               // The current node is already in prevNodes.
               node = prevNodes.back();
               break;
            }
            prevNodes.push_back(node);
         } while (!gotBranch && --dummyCount);
         ASSERT(dummyCount);    // Incorrect termination condition?
         if (node) {
            const Rf_WireContainer wires(node.GetWires());
            Rf_Cell cell;
            Rf_Cell prevCell;
            for (Rf_WireContainer::const_iterator wIt(wires.begin());
                 wIt != wires.end();
                 ++wIt)
            {
               const Rf_Wire wire(*wIt);
               cell = wire.GetCell();
               if (cell != prevCell) {
                  if (!isSafePin && IsInterconnectCell(cell)) {
                     cellFn(cell);
                  }
                  prevCell = cell;
               }
            }
         }
      }
   }
   template <typename CellFn>
   void MapIntTiles(CellFn& cellFn,
                    const Rf_Cell site,
                    const Rf_CellDef& tileDef) const {
      const bool is7series(Is7Series());
      const std::string siteType(site.GetCellDef().GetName());
      if ((siteType == "PAD" || siteType == "IOBM" || siteType == "IOBS")
          && IsSpartan6())
      {
         // Do nothing. Interconnect for Spartan-6 I/Os are not
         // reported correctly, so this redundant check is skipped for
         // Spartan-6.
         ;
      } else {
         if (is7series && IsPrefix("MMCM", siteType.c_str())) {
            // Special case for MMCM. Add 16 interconnect blocks above
            // the top block found by tracing.
            const size_t mmcmExtraCount(16);
            TraceToInt(cellFn, site);
            BoundInts bb;
            TraceToInt(bb, site);
            size_t yTop(bb.top.GetTileRow());
            size_t xLeft(bb.top.GetTileColumn());
            size_t desiredRow(yTop - 1);

            // Ouch time. No direct way to look up a cell at a
            // particular (x, y) offset at this layer of the system.
            // We observe that cell indices are in column-major order,
            // so incrementing by one decreases the row by one.
            Rf_CellDef tcdt(device->GetTopCellDefTile());
            // Add INTs below based on schematic analysis
            for (size_t i(bb.top.GetTileIndex() + 1);
                 i < bb.top.GetTileIndex() + mmcmExtraCount;
                 ++i)
            { 
               Rf_Cell c(tcdt.GetCell(i));
               ASSERT(c);
               ASSERT(c.GetTileColumn() == xLeft);
               ASSERT(c.GetTileRow() == desiredRow);
               ASSERT(IsInterconnectCell(c));
               cellFn(c);
               --desiredRow;
            }
         } else if (is7series && IsPrefix("PLL", siteType.c_str())) {
            // Special case for PLL. Add 13 interconnect blocks below
            // the top block found by tracing.
            const size_t pllExtraCount(13);
            TraceToInt(cellFn, site);
            BoundInts bb;
            TraceToInt(bb, site);
            size_t yBottom(bb.bottom.GetTileRow());
            size_t xLeft(bb.bottom.GetTileColumn());
            size_t desiredRow(yBottom + 1);

            // Ouch time again. No direct way to look up a cell at a
            // particular (x, y) offset at this layer of the system.
            // We observe that cell indices are in column-major order,
            // so decrementing by one increases the row by one.
            Rf_CellDef tcdt(device->GetTopCellDefTile());
            // Add INTs below based on schematic analysis
            for (size_t i(bb.bottom.GetTileIndex() - 1);
                 (bb.bottom.GetTileIndex() - pllExtraCount) < i;
                 --i)
            { 
               Rf_Cell c(tcdt.GetCell(i));
               ASSERT(c);
               ASSERT(c.GetTileColumn() == xLeft);
               ASSERT(c.GetTileRow() == desiredRow);
               ASSERT(IsInterconnectCell(c));
               cellFn(c);
               ++desiredRow;
            }

         } else if (is7series &&
                    (siteType == "IN_FIFO" || siteType == "OUT_FIFO"))
         {
            // Expand the I/O FIFO by one tile if it is adjacent to a
            // PMV. 
            //
            // The PMV tiles occur in pairs. The interconnect from one
            // gets allocated to the fifo above; the interconnect from
            // the other gets allocated to the fifo below. Scan up,
            // stopping when any of the following occur:
            //    * The Top of column is found - no work to do
            //    * An I/O FIFO is found - no work to do
            //    * A PMV is found - Add one interconnect.
            // Same procedure in the down direction, changing top of
            // column to bottom of column.
            Rf_Cell fifoTile(GetSiteParentTile(site));
            TraceToInt(cellFn, site);
            Rf_CellDef tcdt(device->GetTopCellDefTile());
            size_t fifoIndex(fifoTile.GetTileIndex());
            std::string fifoType(fifoTile.GetCellDef().GetName());

            Rf_Cell above;
            size_t aboveIndex(fifoIndex);
            std::string aboveType;
            size_t fifoColumn(fifoTile.GetTileColumn());
            do {
               ++aboveIndex;
               above = tcdt.GetCell(aboveIndex);
               aboveType = std::string(above.GetCellDef().GetName());
            } while (aboveType == "NULL" && above.GetTileColumn() == fifoColumn);

            Rf_Cell below;
            size_t belowIndex(fifoIndex);
            std::string belowType;
            do {
               --belowIndex;
               below = tcdt.GetCell(belowIndex);
               belowType = std::string(below.GetCellDef().GetName());
            } while (belowType == "NULL" && below.GetTileColumn() == fifoColumn);

            BoundInts bb;
            TraceToInt(bb, site);

            if (IsPrefix("CMT_PMV", aboveType.c_str())) {
               Rf_Cell pmvInt(tcdt.GetCell(bb.bottom.GetTileIndex() + 1));
               ASSERT(pmvInt.GetTileColumn() == bb.left.GetTileColumn());
               cellFn(pmvInt);
            }

            if (IsPrefix("CMT_PMV", belowType.c_str())) {
               Rf_Cell pmvInt(tcdt.GetCell(bb.top.GetTileIndex() - 1));
               ASSERT(pmvInt.GetTileColumn() == bb.left.GetTileColumn());
               cellFn(pmvInt);
            }
         } else if (Is7SeriesClockSite(site)) {
            TraceToInt(cellFn, site);
         } else if (IsConfigCenterType(siteType)) {
            TraceToInt(cellFn, site);
         } else if (is7series && !IsPrefix("SLICE", siteType.c_str()) &&
                    IsUserTileType(siteType))
         {
            TraceToInt(cellFn, site);
         } else if (IsZynq() && siteType.c_str() == "T_TERM_INT") {
            ; // Not an interconnect tile. Do nothing.
         } else if (helpers->SpansMultipleInterconnectTiles(site)) {
            typedef const std::vector<Port::INT32> ConstVecInt32;
            ConstVecInt32& iv(helpers->GetInterconnectTilesForCell(site));
            for (ConstVecInt32::const_iterator ivIt(iv.begin());
                 ivIt != iv.end();
                 ++ivIt)
            {
               Rf_Cell intTile(tileDef.GetCell(*ivIt));
               if (intTile) {
                  if (!IsInterconnectCell(intTile)) {
                     const std::string intType(intTile.GetCellDef().GetName());
                     if (intType != "T_TERM_INT") {
                        Warn(145, "IsInterconnectCell() does not recognize " +
                             std::string(intTile.GetName()) + "[" +
                             intType + "]");
                     }
                  }
                  cellFn(intTile);
               }
            }
         } else {
            Port::INT32 intId(helpers->GetInterconnectTileForCell(site));
            Rf_Cell intTile(tileDef.GetCell(intId));
            if (intTile) {
               if (!IsInterconnectCell(intTile)) {
                  const std::string intType(intTile.GetCellDef().GetName());
                  if (intType != "T_TERM_INT") {
                     Warn(145, "IsInterconnectCell() does not recognize " +
                          std::string(intTile.GetName()) + "[" +
                          std::string(intTile.GetCellDef().GetName()) + "]");
                  }
               }
               cellFn(intTile);
            }
         }
      }
   }
   bool SpansMultipleTiles(const Rf_Cell site) const;
   Rf_Cell GetIntTile(const Rf_Cell site,
                      const Rf_CellDef& tileDef,
                      int directionSign) const;
   const Rf_Cell GetSite(const Rf_CellDef& topCellDef,
                         const std::string siteName) const;
   void GetAreaRangeCorners(SiteRect& siteRect);
   bool ParseRangeString(SiteRect& siteRect,
                         const std::string& areaGroupName,
                         const std::string& rangeString);
   bool IsVirtex4() const;
   bool IsVirtex5() const;
   bool Is7Series() const;
   bool IsZynq() const;
   bool IsSpartan6() const;
private:
   const PM_PIC* pmpic;
   bool isPicOwner;
   const Rf_Device* device;
   const Rf_Package* package;
   Rf_TilePinHelpers* helpers;
};

const std::string SetToString(const DeviceModel::CellSet cs,
                              const std::string separator = ", ");

#endif
