// Copyright (C) 2009-2013 Xilinx, Inc.  All rights reserved
// $Id: TileChecker.h,v 1.16.206.17 2013/12/10 19:57:38 jcorbett Exp $
//
// Map the nets to the tile grid. Create contours based on
// isolation groups. Report tiles that contain nets from multiple
// groups. Report tiles that contain isolated nets and are adjacent to
// tiles that contain isolated nets from a distinct isolation group.

#ifndef Iv_TileChecker_h
#define Iv_TileChecker_h

#include <map>
#include <set>
#include <vector>

#include "TUtilities/export/smart_ptr.hpp"
#include "TUtilities/export/regex.hpp"
#include "Portability/export/Port_Type.h"
#include "Databases/Rf/export/Facade/Rf_CellDef.h"

#include "CadFile.h"
#include "ContourMap.h"
#include "Data.h"
#include "DeviceModel.h"
#include "DiagramEmitter.h"
#include "GroupSetFamily.h"
#include "Tile.h"

typedef std::set<std::string> StringSet;
typedef StringSet::iterator StringSetIt;

typedef std::map<std::string, Tile*> SiteNameTileMap;
typedef SiteNameTileMap::iterator SiteNameTileMapIt;

typedef std::map<std::string, SiteNameTileMap> CohortGroupMap;
typedef CohortGroupMap::iterator CohortGroupMapIt;

typedef std::map<Tile*, DeviceModel::CellSet> TileCellSetMap;
typedef TileCellSetMap::iterator TileCellSetMapIt;

class BoundIntsLess {
public:
   bool operator()(const BoundInts& a,
                   const BoundInts& b) const
   {
      if (a.site.GetIndex() == b.site.GetIndex()) return false;
      if (a.bottom.GetTileRow() < b.bottom.GetTileRow()) return true;
      if (a.bottom.GetTileRow() > b.bottom.GetTileRow()) return false;
      if (a.top.GetTileRow() < b.top.GetTileRow()) return true;
      if (a.top.GetTileRow() > b.top.GetTileRow()) return false;
      if (a.site.GetIndex() < b.site.GetIndex()) return true;
      return false;          // Not reached
   }
};

typedef std::set<BoundInts, BoundIntsLess> BoundSet;

// All the interconnect intervals for sites in a given column indexed
// by the minimum interval row coordinate.
typedef std::map<size_t, BoundSet> CmtColumnMap;

// Collection of column records indexed by column.
typedef std::map<size_t, CmtColumnMap> CmtMap;

class TileChecker {
public:
   enum {Unused, Mixed, Ignored, Violation, Overflow, User };
public:
   TileChecker(Data& data);
   ~TileChecker();
   void Check();
   void ReportFaults();
   void ReportAreaSeparation();
   void ReportSpecialTileFaults();
   void GenerateDiagram();
   int GetAreaRangeFaultCount();
   int GetErrorCount();
   int GetStatus();
private:
   // Mapping
   void MapOneWireTile(Rf_Node rfNode,
                       Rf_Wire wire,
                       const std::string& netName,
                       const std::string& Netgroup,
                       XCad_CadNetwork* net);
   void MapIsolationGroupToTiles(const std::string& group);
   void MapNetsToTiles();
   void MapSitesToTiles();
   void MapHiddenStripes();
   void AllocateSingletonGroups();
   void LabelGroupSets();
   bool IsHiddenTile(size_t x, size_t y);
   bool OutsideIsolatedRanges(size_t i);
   void MarkComps();
   void CollectNominalFenceTiles();
   void CollectSuperSiteTileCoordinates();
   void RemoveCells(BoundInts& bb);
   void AddNeighborsForSuperSites();
   void StepAlongPerimeter(Tile* superSite,
                           size_t iPrev,
                           Direction dirPrev,
                           size_t& iNext,
                           Direction& dirNext);
   void AddCorner(size_t index,
                  Direction dirPrev,
                  Direction dirNew,
                  double inset,
                  CoordVec& boundary);
   void AddNeighbor(size_t otx, size_t oty, Tile* superSite);
   void AddNeighborsForRectangle(BoundInts& bb,
                                 const Rf_CellDef& tileDef,
                                 const Rf_Cell superSite);
   void TraceContour(Rf_Cell site,
                     Rf_CellDef tileDef);
   void CollectPolygonalBoundaries();

   // Checking
   size_t CheckSpartan6ClockSite(Tile* tile);
   size_t CheckS6SpecialSite(Tile* tile);

   void Find7SeriesConfigCenterBounds(Port::UINT32& columnMin,
                                      Port::UINT32& rowMin,
                                      Port::UINT32& columnMax,
                                      Port::UINT32& rowMax);
   Port::UINT32 FindLeftInterconnect(Port::UINT32 column, Port::UINT32 row);
   Port::UINT32 FindRightInterconnect(Port::UINT32 column, Port::UINT32 row);
   void Gather7SeriesConfigCenterGroups(Port::UINT32 configIntColumn,
                                        Port::UINT32 rowMin,
                                        Port::UINT32 rowMax);
   void Add7SeriesConfigCenterNeighbor(Port::UINT32 column, Port::UINT32 row);
   void Check7SeriesConfigCenter();
   void CheckZynqPSS();

   void Check7SeriesGTXE2();
   void CheckSpecialSites();
   void CheckPS7Constraint();
   void AddRangeViolation(const Data::IsolationGroupName& g1,
                          const DeviceModel::SiteRect& rc1,
                          const Data::IsolationGroupName& g2,
                          const DeviceModel::SiteRect& rc2,
                          const std::string& violation);
   void CheckAdjacentAreaRanges(Tile* t1, Tile* t2);
   void CheckOverlappingAreaRanges(Tile* tile);
   void MapCmtCohort(const std::string group, Tile* tile, Rf_Cell site);
   void MakeCmtMap(CmtMap& cmtMap);
   Tile* GetKeyTile(Tile* tile);
   void InsertImpliedPhaserSites(SiteNameTileMap&,
                                 const std::string& xStr,
                                 size_t n,
                                 Tile* tile);
   const std::string ComposeImpliedViolation(
      TileSet sourceTiles,
      const std::string group,
      DeviceModel::CellSet impliedSet) const;
   void CheckImpliedIOFifos(CmtMap& cmtMap);
   void CheckImpliedAreaRanges();
   void CheckAreaRanges();
   void CheckTaints();
   void SweepSpecialColumn(Rf_Cell site,
                           StringSet& columnFaultSet,
                           bool isDouble,
                           size_t excludeMin,
                           size_t excludeMax);
   void SweepConfigCenterColumn(Rf_Cell site, bool isAdjacentToClockBuffers);
   void SweepClockBufferColumn(Rf_Cell site);
   void SweepSuperSites();
   void SweepAllAdjacent();
   void CheckTerminals();

   // Reporting
   void ReportMixedTiles();
   void ReportFenceNets();
   void ReportFenceComps();
   void ReportFencePips();
   void ReportAdjacentTiles();
   void ReportConfigCenterFaults();
   void ReportSpecialTiles();
   void ReportSummary();
private:
   Data& data;
   DeviceModel* deviceModel;
   GroupSetFamily groupSetFamily;
   TileVec grid;
   int sweepState;
   DiagramEmitter diagramEmitter;
   // This needs to be a pointer because the grid dimensions are not
   // known at construction time. It is a shared_ptr so that its
   // destructor will be called when the TileChecker is destroyed.
   boost::shared_ptr<ContourMap<Tile> > contourMap;
   IndexVec fenceList;
   // Tile adjacency errors
   TileFaultList netAdjacencyFaults;
   size_t netAdjacencyFaultCount;
   size_t areaRangeFaultCount;
   size_t logicAdjacencyFaultCount;
   size_t netContentFaultCount;
   size_t logicContentFaultCount;
   size_t configCenterFaultCount;
   size_t dspFaultCount;
   size_t impliedSiteCount;
   size_t topLevelCmtCount;
   size_t superFaultCount;
   size_t gtxFaultCount;
   size_t irsNetPipCount;
   size_t irsMultiLoadCount;
   size_t clockFaultCount;
   size_t configColumnFaultCount;
   StringSet rangeViolationSet;
   StringSet s6DspFaultSet;
   StringSet superFaultSet;
   PointSet superSiteSet; // supersite coordinates
   DeviceModel::CellSet superIntSet;
   BoundaryMap superBoundaryMap;
   BoundaryMap superGlobalMap;
   StringSet impliedSiteViolationSet;
   StringSet topLevelCmtSet;
   StringSet gtxFaultSet;
   Data::GroupSet* configCenterGroupSet;
   Data::GroupSet* configCenterNeighborGroupSet;
   TileSet configCenterSet;
   TileSet configCenterNeighborSet;
   TileSet configCenterFaultSet;
   size_t nominalFenceNets;
   bool sweptConfigCenterColumn;
   bool sweptClockBufferColumn;
   StringSet clockFaultSet;
   StringSet configColumnFaultSet;
   size_t configCenterRowMin;
   size_t configCenterRowMax;
};

#endif
