// Copyright (C) 2006-2013 Xilinx, Inc.  All rights reserved
// $Id: CadFile.c,v 1.64.206.8 2014/12/17 20:03:39 jcorbett Exp $
#include <stdexcept>
#include <algorithm>

#include "Xdm/export/xdm.h"
#include "Portability/export/Port_MsgDispatcher.h"
#include "Portability/export/basutstr.h"

#include "main.h"
#include "CadFile.h"

// R#accept_isolation_regions, R#report_range_constraints

using namespace std;

// required for Port Utilities
static const char modname[] = __FILE__;
static const char rcsrev[] = "$Revision: 1.64.206.8 $";

// Single NCD loading

// Returns true when the given block is directly below the top level
// and contains blocks.
struct ImplementedModuleFilter {
   bool operator()(const XCad_CadBlock* block) {
      if (block->GetHierDepth() == 1 && 0 < block->GetNumBlocks()) {
         if (block->GetPropertyByName("USER_VISIBLE")) {
            return true;
         }
      }
      return false;
   }
};

// R#accept_isolation_regions, R#report_range_constraints - Returns true
// when the given block is below the top level and has the name given
// in the constructor.
struct NamedModuleFilter {
   // Initialize the filter.
   NamedModuleFilter(const string& name) : name(name) {}
   bool operator()(const XCad_CadBlock* block) {
      string blockPath(block->GetUserHierName());
      return name == blockPath;
   }
   const string name;
};

// Both single and multiple NCD loading

// Load a placed and routed FPGA design file.
void CadFile::LoadNcd(const string& ncdPath) {
   try {
      SuppressLibraryMessages suppressor;
      loader = new Pds_PahDesignLoaderT<XCad_CadTraits>();
      view = loader->LoadAndSpecialize(ncdPath.c_str(),
                                       APPNAMESTR,
                                       Xdm_Txn::READ_ONLY);
      rmed = Util::SharedObject<XCad_RMed>::Key(view).Get();
      if (!rmed) {
         rmed = Util::SharedObject<XCad_RMed>::Key(view).Create(view);
      }
      // If view were invalid, attempts to get or create the resource
      // mediator would fail.
   } catch (Pds_Exception& e) { // R#report_errors
      const string fileName(GetFileName());
      // UT_STR::LstrToMbstr() uses new char[], but we're about to
      // exit so we will leak...
      const string errorMessage(
         "Input Error:7: Failed to load design "
         + fileName + ".\n"
         + string(UT_STR::LstrToMbstr(e.getLocalized())));
      Console("\n" + errorMessage);
      Finalize();
      throw runtime_error(errorMessage);
   } catch (...) {              // R#report_errors
      const string fileName(GetFileName());
      const string errorMessage("Input Error:54: Failed to load design "
                                + fileName + ".");
      Console("\n" + errorMessage);
      Finalize();
      throw runtime_error(errorMessage);
   }
}

// R#accept_isolation_regions, R#report_range_constraints - This
// constructor is used for the combined design with all placement and
// routing for the system to be verified.
//
// The loader is held because the view is invalidated when it is destroyed.
// The rmed is held because it has a funky allocator and deallocator.
CadFile::CadFile(const string& ncdPath)
   : ncdPath(ncdPath.c_str()), module(0), loader(0), deviceModel(0),
     isDeviceModelOwner(false), view(0), rmed(0)
{
   Console("Loading " + GetFileName() + "...");
   LoadNcd(ncdPath);

   // We need all the nets, not only the routable ones returned
   // by rmed->GetRoutableNetworks(nets);

   XCad_CadView::Networks cadNets(view->GetNetworks());
   nets.reserve(view->GetMaxNetworks());
   for (XCad_CadView::Networks::iterator cnIt(cadNets.begin());
        cnIt != cadNets.end();
        ++cnIt)
   {
      XCad_CadNetwork* net(*cnIt);
      nets.push_back(net);
   }
   Console("done.\n");
}

void CadFile::SetDeviceModel(DeviceModel* deviceModel) {
   this->deviceModel = deviceModel;
}

// R#accept_isolation_regions
DeviceModel* CadFile::GetDeviceModel() {
   ASSERT(deviceModel);
   return deviceModel;
}

// R#accept_isolation_regions - Is the given net in the block hierarchy
// of this CadFile?
bool CadFile::HasNet(const XCad_CadNetwork* net,
                     const XCad_CadBlock* module) const
{
   bool result(false);
   const XCad_CadSignal* sig(net->GetTopSignal());
   for (const XCad_CadBlock* parent(sig->GetParent());
        parent;
        parent = parent->GetParent())
   {
      if (module == parent) {
         result = true;
         break;
      }
   }
   return result;
}

// Is this net in one of the isolated blocks?
bool CadFile::IsIsolatedNet(const std::vector<XCad_CadBlock*>& isolatedModules,
                            const XCad_CadNetwork* net) const
{
   bool isIsolated(false);
   typedef const std::vector<XCad_CadBlock*> Blocks;
   for (Blocks::const_iterator imIt(isolatedModules.begin());
        imIt != isolatedModules.end() && !isIsolated;
        ++imIt)
   {
      const XCad_CadBlock* block(*imIt);
      isIsolated = isIsolated || HasNet(net, block);
   }
   return isIsolated;
}

// R#report_errors - If a module name does not correspond to a unique
// user-visible block, this method reports the error and lists valid
// block names.  Note: this method only lists modules directly below
// top, so as to avoid listing every user visible block at all levels
// below top.
void CadFile::ReportInvalidModule(vector<XCad_CadBlock*>& blocks,
                                  const string& ncdOrModule,
                                  boost::shared_ptr<CadFile>& combined)
{
   ImplementedModuleFilter allModules;
   Port::UINT32 count(view->GetFilteredBlocks(blocks, allModules));
   if (count == 0) {      // R#report_errors
      const string errorMessage(
         "Input Error:73: No user visible modules were found in "
         + combined->GetFileName() + ".\n"
         "Check that the correct NCD file was supplied.");
      Console("\n" + errorMessage);
      throw runtime_error(errorMessage);
   } else {               // R#report_errors
      string sep("  ");
      string modList;
      for (vector<XCad_CadBlock*>::iterator bIt(blocks.begin());
           bIt != blocks.end();
           ++bIt)
      {
         XCad_CadBlock* block(*bIt);
         modList += sep + string(block->GetName());
         sep = ", ";
      }
      const string errorMessage(
         "Input Error:74: '" + ncdOrModule + "' is not a module name in "
         + combined->GetFileName() + ".\n"
         "The modules found were:\n" + modList);
      throw runtime_error(errorMessage);
   }
}

// R#accept_isolation_regions - This constructor is used for
// modules. It may be given an NCD file path or the name of a module
// in the combined design.  If the first parameter ends in ".ncd" it
// is assumed to be an NCD file path.
CadFile::CadFile(const string& ncdOrModule,
                 boost::shared_ptr<CadFile>& combined,
                 const std::vector<XCad_CadBlock*>& isolatedModules)
   : ncdPath(""), module(0), loader(0), deviceModel(0),
     isDeviceModelOwner(false), view(0), rmed(0)
{
   Port_MsgDispatcher::getDefaultMsgMan()->enable();
   ncdPath = ncdOrModule.c_str();
   if (ncdPath.GetExtension() == ".ncd") {
      // Assume ncdOrModule is a NCD file path, not a
      // design element named something.ncd.

      Console("Loading " + GetFileName() + "...");
      LoadNcd(ncdOrModule);
      deviceModel = new DeviceModel(view->GetDevice());
      isDeviceModelOwner = true;

      // Find the module name. We're looking for the block
      // directly under /top/PACKED/top that has the USER_VISIBLE
      // property and at least 1 child block. There should be just
      // one unless this is the combined design.

      // GetFilteredBlocks examines every block in the
      // hierarchy. Would be more efficient to use GetBlocks(),
      // but it would take a little more code.
      ImplementedModuleFilter allModules;
      vector<XCad_CadBlock*> blocks;
      Port::UINT32 count(view->GetFilteredBlocks(blocks, allModules));
      if (count != 1) {         // R#report_errors
         const string errorMessage(
            "Input Error:72: Group NCD file " + ncdOrModule + " has more\n"
            "than one implemented block below top.\n"
            "An NCD file that defines an isolation group cannot be\n"
            "merged nor complete. Consider supplying the module name\n"
            "instead of an NCD file path.");
         Console("\n" + errorMessage);
         throw runtime_error(errorMessage);
      }

      module = blocks.front();
      Console("Found module " + GetModuleName()
              + " in " + ncdOrModule + ".\n");

      // We need all the nets, not only the routable ones returned
      // by rmed->GetRoutableNetworks(nets);
      XCad_CadView::Networks cadNets(view->GetNetworks());
      nets.reserve(view->GetMaxNetworks());
      for (XCad_CadView::Networks::iterator cnIt(cadNets.begin());
           cnIt != cadNets.end();
           ++cnIt)
      {
         XCad_CadNetwork* moduleNet(*cnIt);
         nets.push_back(moduleNet);
      }

      Console("done.\n");
   } else {
      // Assume ncdOrModule is a module name, not an
      // NCD file path.
      ncdPath = combined->GetFileName().c_str();
      view = combined->GetView();
      deviceModel = combined->GetDeviceModel();
      rmed = combined->GetRMed();

      NamedModuleFilter isMod(ncdOrModule);
      vector<XCad_CadBlock*> blocks;
      Port::UINT32 count(view->GetFilteredBlocks(blocks, isMod));
      if (count != 1) {
         // Report a fatal error and exit.
         ReportInvalidModule(blocks, ncdOrModule, combined);
      }
      module = blocks.front();
      Console("Found module " + ncdOrModule
              + " in " + combined->GetFileName()
              + ".\n");

      // We need all the nets, not only the routable ones returned
      // by rmed->GetRoutableNetworks(nets);
      XCad_CadView::Networks cadNets(view->GetNetworks());
      nets.reserve(view->GetMaxNetworks());
      for (XCad_CadView::Networks::iterator cnIt(cadNets.begin());
           cnIt != cadNets.end();
           ++cnIt)
      {
         XCad_CadNetwork* net(*cnIt);

         // To match the old behavior, we need all the nets in the
         // module and all the nets not in any module.
         if (!IsIsolatedNet(isolatedModules, net) || HasNet(net, module)) {
            nets.push_back(net);
         }
      }
   }
}

void CadFile::Finalize() {
   try {
      if (loader) {
         Util::SharedObject<XCad_RMed>::Key(view).Release(rmed);
         delete loader;
         loader = 0;
      }
      if (isDeviceModelOwner) {
         delete deviceModel;
         isDeviceModelOwner = false;
      }
      deviceModel = 0;
      rmed = 0;
      view = 0;
   } catch (...) {
      ; // Nothing more we can do here. Might already be processing an
        // exception.
   }
}

CadFile::~CadFile() {
   Finalize();
}


// R#accept_isolation_regions - Return the module name or the empty
// string.
const string CadFile::GetModuleName() {
   return string(module ? module->GetReportableName() : "");
}

// R#report_errors - Return the part after the final slash.
const string CadFile::GetFileName() {
   return string(ncdPath.GetTail());
}

// R#accept_isolation_regions - The module block gives us access to
// various aspects of the module such as the name, the nets, the
// ports, etc.
XCad_CadBlock* CadFile::GetModule() {
   return module;
}

// R#accept_isolation_regions - Since module blocks may not be directly
// below top, this provides a mechanism to look up a block by the path
// given on the command line. User global nets are nets outside of all
// isolated blocks; not necessarily at the top level of the hierarchy.
XCad_CadBlock* CadFile::GetBlockByUserHierName(const std::string& name) {
   return view->GetBlockByUserHierName(name.c_str());
}


// R#accept_isolation_regions - The view gives us access to the blocks
// in the design and also to the part for which the design was
// generated.
XCad_CadView* CadFile::GetView() { return view; }

// R#accept_isolation_regions - The rmed (resource mediator) gives us
// access to the nodes in nets in the design.
XCad_RMed* CadFile::GetRMed() { return rmed; }

// R#accept_isolation_regions - Get all the nets in the module or design.
CadFile::NetVec& CadFile::GetNets() { return nets; }

// R#report_range_constraints - Matches a list of ranges separated by
// commas. Marked subexpressions enable capture of the coordinates of
// the first range, and the rest ranges.
const string num("(\\d+)");       // a non-negative integer
const string site("([A-Z0-9_]+)");
const boost::regex
rangeListPat(site + "_X" + num + "Y" + num
             + ":\\1_"
             "X" + num + "Y" + num
             + "(?:,(.*))?",
             boost::regex::normal | boost::regbase::icase);

// R#accept_isolation_regions, R#report_range_constraints - Parse a
// string describing one or more rectangles into a list of SiteRect
// structures. Returns true if the entire string was successfully
// consumed by repeated applications of the pattern and false
// otherwise. Stores the results in siteRectList.
bool CadFile::ParseOneRangeList(const string& areaGroup,
                                const string& rangeListString)
{
   ASSERT(deviceModel);
   boost::smatch itemResult;
   string rs(rangeListString);
   while (boost::regex_match(rs, itemResult, rangeListPat)) {
      DeviceModel::SiteRect siteRect;
      // Length of optional ',' + subsequent ranges.
      if (itemResult.str(6).empty()) {
         siteRect.rangeString = itemResult.str(0);
      } else {
         size_t suffixLength(1 + itemResult.str(6).size());
         siteRect.rangeString = rs.substr(0, rs.size() - suffixLength);
      }
      bool gotOne(deviceModel->ParseRangeString(siteRect,
                                                areaGroup,
                                                siteRect.rangeString));
      if (gotOne) {
         if (siteRect.lowerInt && siteRect.upperInt) {
            // Collect parsed range.
            siteRectList.push_back(siteRect);
         } else {
            // For 7 series, we'll get here for sites like BUFIO,
            // CFG_IO_ACCESS, and IBUFDS.
         }
      } else {
         ;
         // Got something like STARTUP or FRAME_ECC. Must rely
         // on manual analysis for now.
      }
      // Discard processed text.
      rs = itemResult.str(6);
   }
   sort(siteRectList.begin(), siteRectList.end(),
        DeviceModel::SiteRectLess());
   return rs.empty();            // Consumed all.
}

// R#accept_isolation_regions, R#report_range_constraints - Find the
// area range constraints associated with this module.
//
// Each constraint is stored in a string-valued property an object in
// the data model. The string value contains a semi-colon separated
// list of attribute=value pairs. The objects the properties are
// associated with are not the module blocks that represent the user's
// design.
//
// Circa ISE 9.2.04, the object with the property can be obtained from
// an Xdm binding that associates the design module under
// /<model>/PACKED/<model>/<module> with the area group under
// /<model>/IST_HIERARCHY_HD/<model>/<area_group>.
//
// Circa ISE 11.1, the property name is stored as one of the attribute
// values within the property value. The PARTITION portion of the
// property value is the module name. The value matches the pattern
//
//    (^|;)PARTITION=/top/<module_name>;
//
// The area range we seek matches the pattern
//
//    SLICE[LM]_RANGE=SLICE_X[0-9]+Y[0-9]+(,SLICE_X[0-9]+Y[0-9]+)*;
//
void CadFile::FindSiteRangeString() {
   const boost::regex
      partitionPat("(?:.*[^\\\\];)?"       // semicolon-delimited attributes 
                   "PARTITION="            // partition attribute
                   "/[^/]+/"               // top
                   "((?:(?:\\\\;)|[^;])+)" // partition name (captured)
                   ".*"                    // additional attributes
                   );
   const string moduleName(module->GetUniqueHierName());
   string errorCode("");

// The commented line below is the only difference required to port
// back to ISE 11.x.
// //   Xdm_Model* xdmModel(view->GetNativeXdmModel());
   Xdm_Model* xdmModel(view->GetModel());
   ASSERT(xdmModel);            // Extremely unlikely to fail...

   Xdm_DesignMgr* mgr(xdmModel->GetDesignMgr());

   const string vn("/" + string(mgr->GetName()) + "/IST_HIERARCHY_HD");
   Xdm_View* xdmView(xdmModel->GetViewByUniqueHierName(vn.c_str()));
   if (xdmView) {
      Xdm_Block* xdmTop(xdmView->GetBlock());
      if (xdmTop) {
         // One of the modules (blocks below top) has the needed
         // information on a property. To identify the relevant
         // object, we need to either find the module name in the
         // property value or find a binding named after the module.
         Xdm_Block::Blocks moduleList(xdmTop->GetBlocks());
         bool foundBlock(false);
         for (Xdm_Block::Blocks::iterator bIt(moduleList.begin());
              bIt != moduleList.end();
              ++bIt)
         {
            Xdm_Block* moduleBlock(*bIt);
            if (moduleBlock) {
               // The moduleBlock corresponds to one of the user's
               // area groups.
               const char* pn("IST_ATTRIBUTE_PERSISTENT_PROP_STRING");
               Xdm_Property* xdmProp(moduleBlock->GetPropertyByName(pn));
               if (xdmProp) {
                  string propValue(xdmProp->GetPropertyString());
                  // The partition attribute may be first, middle, or
                  // last. The module name may be a suffix of the
                  // partition name or a delimited section in the
                  // middle. We shouldn't mistake /top/foo as a match
                  // for /top/foobar, however /top/foo should match
                  // /top/foo/bar.
                  //
                  // Note: assumes backslash (\) may be used as an
                  // escape character to embed forward slashes and
                  // semicolons in names.
                  boost::smatch result;
                  bool foundPartition(boost::regex_match(propValue,
                                                         result,
                                                         partitionPat));
                  if (foundPartition) {
                     // The property name matches the module name if
                     // either the two are identical or the module
                     // name is a properly delimited suffix of the
                     // property name.

                     const string pName(result.str(1));
                     const string mName(GetModuleName());
                     size_t pz(pName.size());
                     size_t mz(mName.size());
                     if (pName.find(mName) == 0) {
                        // mName is a prefix of pName so pName[mz] is defined.
                        if (mz == pz || pName.at(mz) == '/') {
                           foundBlock = true;
                           ParseRangeList(moduleBlock->GetName(), propValue);
                        }
                     }
                  } else {
                     // Try to get the area group from the bindings
                     // need to get from moduleBlock,
                     // e.g. /top/IST_HIERARCHY_HD/top/AG_region_sw to
                     // the corresponding module block in the
                     // /top/PACKED hierarchy.
                     const string
                        bn("ISET_MEMBER_BLOCK_BINDING/"
                           + string(mgr->GetName())
                           + "/IST_HIERARCHY_HD");
                     Xdm_Binding*
                        binding(moduleBlock->GetBindingByName(bn.c_str()));
                     if (binding) {
                        Xdm_Binding::BindingObjects
                           bos(binding->GetBindingObjects());
                        // Exactly two binding objects are
                        // expected. One named after the module. The
                        // other named after the area group.
                        for (Xdm_Binding::BindingObjects::const_iterator
                                boIt(bos.begin());
                             boIt != bos.end();
                             ++boIt)
                        {
                           Xdm_Binding::BindingObjects::value_type v(*boIt);
                           string bindingName(v->GetUniqueHierName());
                           if (bindingName == moduleName) {
                              foundBlock = true;
                              ParseRangeList(moduleBlock->GetName(),
                                             propValue);
                              break;
                           }
                        }
                     }
                  }
               } else {
                  errorCode += "P"; // No property.
               }
            } else {
               errorCode += "M"; // Iterator produced null.
            }
         }
         if (!foundBlock) {
            errorCode += "Z";
         }
      } else {
         errorCode = "T";       // No top block. Malformed design file.
      }
   } else {
      errorCode = "V(" + vn + ")"; // No view. Malformed design file.
   }
   if (!errorCode.empty()) {
      FatalInternal(110, "Unable to find area range data. Code: "
                    + errorCode);
   }
}

// Note the non-greedy *? which matches the shortest possible string.
const boost::regex siteRangeListPat("(.*?;)?[A-Z0-9_]+_RANGE=([^;]+)(;.*)?");

// R#accept_isolation_regions, R#report_range_constraints - Extract
// range coordinates from the given range string and store the parsed
// results in siteRectList.
void CadFile::ParseRangeList(const string& areaGroup,
                             const string& rangeListString) {
   string errorCode;
   string rls(rangeListString);
   boost::smatch result;
   while (boost::regex_match(rls, result, siteRangeListPat)) {
      if (result.size() != 4) {
         FatalInternal(94, "Invalid regular expression.");
      }
      string rangeString(result.str(2));
      bool gotRangeList(ParseOneRangeList(areaGroup, rangeString));
      if (!gotRangeList) {
         errorCode += "E(" + rls + ")"; // Unrecognized range format.
      }
      // Move on to the next.
      rls = result.str(3);
   }
   if (siteRectList.empty()) {
      // Nope. Didn't find any range data.
      errorCode += "S(" + rls + ")";
   }
   if (!errorCode.empty()) {
      FatalInternal(116, "Unable to parse range constraint. Code: "
                    + errorCode);
   }
}

// The test case associated with CR470509 demonstrates that the range
// constraint data can also be stored in a string list property on
// blocks below top. The property is called NPR_CST_RANGES.
//
// There was just one area group name per module and it was stored in
// a property called NPR_CST_ATTRIBUTES. The value would look
// something like this:
//
//    _agName=pblock_aurora_core_aurora_example_1_i,_isRoutingContained=1,
//    _allowRoutingOverBoundary=0,_allowStaticRouting=1,_allowStaticPlacement=1,
//    _plIncludeAreaSites=1,_plExpandAreaSites=0,_plIncludeLockedLogic=0,
//    _rtIncludeAreaRouting=1,_rtExpandAreaRouting=0
// Note: whitespace added for presentation.
//
// The pattern below does not assume attribute=value expressions
// appear in order. It matches an optional list of expressions ending
// with a comma, the expression specifying the area group name, and an
// optional list of expressions.
const boost::regex
nprAttributePat("(?:.*,)?_agName=([^ ,]+)(?:(,.*$))?",
                boost::regex::normal | boost::regbase::icase);

// R#accept_isolation_regions, R#report_range_constraints - Find area
// ranges stored in the NPR_CST_RANGES property and the area group
// name from the NPR_CST_ATTRIBUTES property. This method was used in
// early PR implementations circa ISE 8.x.
pair<const string, const string> CadFile::FindNprSiteRangeListText() {
   string areaGroupName;
   string rangeListString;

   // Get the top level block.
   XCad_CadBlock* top(view->GetBlock());
   ASSERT(top);

   // Fetch the property list.
   const char* ncr("NPR_CST_RANGES");
   Xdm_Property* xdmProp(module->GetPropertyByName(ncr));
   if (xdmProp) {
      // Note: the test case for this method no longer loads so
      // the code below to get the area group name is untested.
      boost::smatch result;
      const char* nca("NPR_CST_ATTRIBUTES");
      Xdm_Property* nprCstAttributes(module->GetPropertyByName(nca));
      if (nprCstAttributes &&
          boost::regex_match(nprCstAttributes->GetPropertyString(),
                             result,
                             nprAttributePat))
      {
         areaGroupName = result.str(1);
      } else {
         FatalInternal(121, "FindNprSiteRangeListText failed to find "
                       "area group name for "
                       "module " + string(module->GetName()) + ".");
      }

      // This property has a string list value. Add each string to
      // the list.
      string sep;
      typedef Xdm_Property::Strings XdmStrings;
      XdmStrings props(xdmProp->GetPropertyStringListItems());
      for (XdmStrings::const_iterator sIt(props.begin());
           sIt != props.end();
           ++sIt)
      {
         rangeListString += sep + *sIt;
         sep = ",";
      }
   }
   return make_pair(areaGroupName, rangeListString);
}

// R#accept_isolation_regions, R#report_range_constraints - Fill the
// siteRectList with all the site ranges and individual sites in the
// design. Individual sites are converted to ranges of one site,
// e.g. "FOO_X0Y0:FOO_X0Y0". Returns a reference to the siteRectList
// of this module. The list is sorted and free of duplicates.
DeviceModel::SiteRectVec& CadFile::GetSiteRectList() {
   if (siteRectList.empty()) {
      pair<string, string> nameAndRange(FindNprSiteRangeListText());
      if (nameAndRange.second.empty()) {
         FindSiteRangeString();
      } else {
         ParseRangeList(nameAndRange.first, nameAndRange.second);
      }
   }
   if (siteRectList.empty()) {
      Fatal(95, "Unable to obtain area range constraints for module "
            + GetModuleName() + "\n"
            "Either this module is not isolated and therefore should not have"
            "\n"
            "associated -group parameters on the command line or "
            "isolation-related\n"
            "design constraints were not defined correctly.");
   }
   // At this point, siteRectList contains all the ranges sorted, but there
   // may be duplicates. The Standard Template Library unique algorithm
   // shuffles siteRect to collect only unique elements in the front
   // portion of the vector. Following this operation, duplicates
   // can be removed with the iterator returned by unique().
   DeviceModel::SiteRectVecIt
      endIt(unique(siteRectList.begin(), siteRectList.end(),
                   DeviceModel::SiteRectEqual()));
   siteRectList.erase(endIt, siteRectList.end());
   return siteRectList;
}

// Is the given tile location part of any of the area range
// constraints attached to this module?
bool CadFile::Contains(Port::UINT32 column, Port::UINT32 row) {
   ASSERT(deviceModel);
   bool result(false);
   if (siteRectList.empty()) {
      GetSiteRectList();
   }
   for (DeviceModel::SiteRectVecIt rIt(siteRectList.begin());
        rIt != siteRectList.end();
        ++rIt)
   {
      const DeviceModel::SiteRect rectangle(*rIt);
      if (rectangle.Contains(column, row)) {
         result = true;
         break;
      }
   }
   return result;
}

// R#accept_isolation_regions - Is the tile at (column, row) contained
// in a rectangular region associated with this module of the
// corresponding site type?
//
// Ensures the siteRectList has been gathered and performs a linear
// search.
bool CadFile::Contains(Port::UINT32 column, Port::UINT32 row,
                       const string& siteType)
{
   ASSERT(deviceModel);
   bool result(false);
   if (siteRectList.empty()) {
      GetSiteRectList();
   }
   for (DeviceModel::SiteRectVecIt rIt(siteRectList.begin());
        rIt != siteRectList.end();
        ++rIt)
   {
      const DeviceModel::SiteRect rectangle(*rIt);
      if (rectangle.Contains(column, row, siteType)) {
         result = true;
         break;
      }
   }
   return result;
}

